/*
**  Copyright (C) 1993/1994, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/

/*
** Telnet Initiator (ti) Main Program
**
** Author: 	Andrew Hewett
**		Universitaet Oldenburg, Germany
** Created:	08/1993
**
** Last Update:		hewett
** Update Date:		1996/05/30 17:41:27
** Source File:		/share/dicom/cvs-depot/dcmtk/imagectn/apps/ti.cc,v
** CVS/RCS Revision:	1.5
** Status:		Exp
**
** CVS/RCS Log at end of file
*/

#include "osconfig.h"    /* make sure OS specific configuration is included first */

#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#include <time.h>
#include <errno.h>
#include <signal.h>

#include "ti.h"
#include "assoc.h"
#include "dimse.h"
#include "cnf.h"
#include "dcdict.h"
#include "dcdebug.h"
#ifndef  NODISPLAY
#include "dispuser.h"
#endif
#include "tiui.h"
#include "tinet.h"
#include "tiquery.h"

#define DEFAULTAPPLICATIONTITLE	"TELNET_INITIATOR"	/* default */


char* progname = NULL;
BOOLEAN verbose = FALSE;
BOOLEAN debug = FALSE;
static char *configFileName = "configrc";
BOOLEAN disableDisplay = TRUE;
BOOLEAN ctnDisplayCmdLineOverride = FALSE;
BOOLEAN displayTransferProgress = TRUE;
BOOLEAN disableImageDisplay = FALSE;

#define MAXREMOTEDBTITLES 20
char *remoteDBTitles[MAXREMOTEDBTITLES];
int remoteDBTitlesCount = 0;

TI_Config conf;


/*
 * My usage stuff
 */ 

static void
shortusage()
{
    fprintf(stderr, "\
usage: %s [-c configrc][-D][-DD][-t title][-v][-d][-I][-P][-R title]\n\
          [-b n] peer ...\n", 
	progname);
}

static void
fullusage()
{
    shortusage();
    fprintf(stderr, "\
parameters:\n\
    peer	peer DPN host name or vendor name (from config file)\n\
options:\n\
    -c configrc use specific configuration file (default: %s)\n\
    -D		enable display\n\
    -DD		disable display\n\
    -t title	my calling AE title (default: %s)\n\
    -v		verbose mode\n\
    -d		debug mode\n\
    -I		disable image display\n\
    -P		do not display image transfer progress\n\
    -R title    remote DB AE title\n\
    -b n	set max receive pdu to n bytes (default: %d)\n",
    configFileName,  DEFAULTAPPLICATIONTITLE, ASC_DEFAULTMAXPDU);
    exit(1);
}

static void 
usage()
{
    shortusage();
    exit(1);
}

void errmsg(const char* msg, ...)
{
	va_list args;
	
	fprintf(stderr, "%s: ", progname);
	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
}

static BOOLEAN 
addPeerName(TI_Config *conf, const char *peerName);
static void
createConfigEntries(TI_Config *conf);
static void 
printConfig(TI_Config *conf);

static void
signalHandler(int sig);

/*
 * Program Entry Point
 */

int 
main(int argc, char *argv[])
{
    CONDITION cond;
    int i;
    char **vendorHosts; 
    char **aeTitleList;
    int maxPDU;

    /* strip any leading path from program name */
    if ((progname = (char *)strrchr(argv[0], PATH_SEPARATOR)) != NULL) {
	progname++;
    } else {
	progname = argv[0];
    }

    if (argc < 2) {
	fullusage();
    }

    bzero((char*)&conf, sizeof(conf));
    conf.myAETitle = DEFAULTAPPLICATIONTITLE;
    conf.maxReceivePDULength = -1;
   
    /* parse program arguments */
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	switch (argv[i][1]) {
	case 'D':
	    disableDisplay = (argv[i][2] == 'D');
	    ctnDisplayCmdLineOverride = TRUE;
#ifdef NODISPLAY
	    fprintf(stderr, "Warning: not compiled for ctndisplay\n");
#endif	    
	    break;
	case 'v':
	    verbose = TRUE;
	    break;
	case 'd':
	    debug = TRUE;
	    verbose = TRUE;
	    break;
	case 'I':
	    disableImageDisplay = TRUE;
	    break;
	case 'P':
	    displayTransferProgress = FALSE;
	    break;
	case 'c':
	    if (i++ < argc) {
		configFileName = argv[i];
	    } else
		usage();
	    break;
	case 't':
	    if (i++ < argc) {
		conf.myAETitle = argv[i];
	    } else
		usage();
	    break;
	case 'R':
	    if (i++ < argc) {
	        if (remoteDBTitlesCount < MAXREMOTEDBTITLES) {
		    remoteDBTitles[remoteDBTitlesCount++] = argv[i];
		} else {
		    errmsg("Too many remote DB titles (max: %d)", 
		        MAXREMOTEDBTITLES);
		    errmsg("Ignored temote DB: %s", argv[i]);
		}
	    } else
		usage();
	    break;
	case 'b':
	    if (((i + 1) < argc) &&
		(sscanf(argv[i + 1], "%d", &conf.maxReceivePDULength) == 1)) {
		i++;		/* Maximum Receive PDU Length parsed */
		if (conf.maxReceivePDULength < ASC_MINIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too small",
			   conf.maxReceivePDULength);
		    usage();
		} else if (conf.maxReceivePDULength > ASC_MAXIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too big",
			   conf.maxReceivePDULength);
		    usage();
		}
	    } else {
		usage();
	    }
	    break;
	default:
	    usage();
	}
    }

    if ((argc - i) < 1) {
	errmsg("Peer Host Name missing");
	usage();
    }

    if ((argc - i) > TI_MAXPEERS) {
	errmsg("Too many peers specified (maximum: %d): %d", 
	    TI_MAXPEERS, (argc - i));
	usage();
    }

#ifdef USE_MALLOC_DEBUG
#ifdef HAVE_MALLOC_DEBUG
#define MALLOCDEBUGLEVEL 16
    if (verbose) {
        printf("malloc_debug on (%d)\n", MALLOCDEBUGLEVEL);
    }
    malloc_debug(MALLOCDEBUGLEVEL);
#endif
#endif

    if (access(configFileName, R_OK) < 0) {
	errmsg("cannot access %s: %s", configFileName, 
		strerror(errno));
	usage();
    }

    if (!CNF_init(configFileName)) {
	errmsg("bad config file: %s", configFileName);
	usage();
    }

    if (verbose) {
	CNF_printConfig();    
    }

    maxPDU = CNF_getMaxPDUSize();
    if (maxPDU == 0) {
        /* not set, set default */
	maxPDU = ASC_DEFAULTMAXPDU;
    }
    
    if (maxPDU < ASC_MINIMUMPDUSIZE || maxPDU > ASC_MAXIMUMPDUSIZE) {
        errmsg("bad MaxPDUSize (%d) in config file: %s", 
	    maxPDU, configFileName);
	usage();
    }

    /* if nothing was set on command line then use config file value */
    if (conf.maxReceivePDULength == -1) {
        conf.maxReceivePDULength = maxPDU;
    }


    /* peer to call */
    while (i < argc) {
        int n;
	if ((n = CNF_aeTitlesForSymbolicName(argv[i], &aeTitleList)) > 0) {
	    int j;
	    const char *peerName = NULL;
	    int portNumber;
	    for (j=0; j<n; j++) {
		if (CNF_peerForAETitle(aeTitleList[j], &peerName, &portNumber)) {
		    addPeerName(&conf, peerName);
		}
	    }
	    if (aeTitleList) free(aeTitleList);
	    aeTitleList = NULL;
        } else if ((n = CNF_HostNamesForVendor(argv[i], &vendorHosts)) > 0) {
	    int j;
	    for (j = 0; j<n; j++) {
	        addPeerName(&conf, vendorHosts[j]);
	    }
	    if (vendorHosts) free(vendorHosts); 
	    vendorHosts = NULL;
	} else {
	    addPeerName(&conf, argv[i]);
	}
        i++;
    }
    conf.peerHostName = conf.peerNames[0];	/* default */

    /* load up configuration info */
    createConfigEntries(&conf);

    if (conf.dbCount < 1) {
	errmsg("no accessable databases");
	usage();
    }

    if (verbose) {
	printConfig(&conf);
    }

    /* set debugging if required */
    DUL_Debug(debug);
    DIMSE_debug(debug);
    SetDebugLevel(((debug)?(3):(0)));	/* dcmdata debugging */

    /* make sure data dictionary is loaded */
    if (dcmDataDict.numberOfEntries() == 0) {
	fprintf(stderr, "Warning: no data dictionary loaded, check environment variable: %s\n",
		DCM_DICT_ENVIRONMENT_VARIABLE);
    }

    /* start up network */
    cond = ASC_initializeNetwork(NET_REQUESTOR, 0, 200, &conf.net);
    if (!SUCCESS(cond)) {
	errmsg("Error initialising network:");
	COND_DumpConditions();
	exit(1);
    }

#ifndef NODISPLAY
    /* initialize display */
    if (!ctnDisplayCmdLineOverride) {
	disableDisplay = (!CNF_getDisplay());
    }
    DSU_setDisplayMode(!disableDisplay);
    if (!DSU_initDisplay(DSU_TIUser)) {
	disableDisplay = TRUE;
    }
#endif

    signal(SIGHUP, signalHandler);
    signal(SIGINT, signalHandler);
    signal(SIGQUIT, signalHandler);
    signal(SIGTERM, signalHandler);

    /*
     * Do the real work
     */
    TI_userInput(&conf);
    

    /* clean up network */
    cond = ASC_dropNetwork(&conf.net);
    if (!SUCCESS(cond)) {
	errmsg("Error dropping network:");
	COND_DumpConditions();
	exit(1);
    }
    return 0;
}

static void 
printConfig(TI_Config *conf)
{
    int i,j;

    printf("TI Configuration:\n");
    printf("My AE Title: %s\n", conf->myAETitle);
    printf("DatabaseTitles    Peer AE Titles\n");
    for (i=0; i<conf->dbCount; i++) {
	printf("%-18s", conf->dbEntries[i]->title);
	for (j=0; j<conf->dbEntries[i]->peerTitleCount; j++) {
	    printf("%s ", conf->dbEntries[i]->peerTitles[j]);
	}
	printf("\n");
    }
}

static BOOLEAN
addPeerName(TI_Config * conf, const char *peerName)
{
    int k;
    BOOLEAN found = FALSE;
    char **aeTitles;

    /* only hosts not already known */
    for (k = 0; !found && k<conf->peerNamesCount; k++) {
	found = (strcmp(conf->peerNames[k], peerName) == 0);
    }
    if (found) {
    	/* already there */
        return TRUE;
    }
    if (CNF_aeTitlesForPeer(peerName, &aeTitles) <= 0) {
        /* no titles defined */
	errmsg("no AE titles defined (in: %s) for peer: %s", configFileName,
	    peerName);
	return FALSE;
    }
    free(aeTitles);	/* do not need aeTitles */

    conf->peerNames[conf->peerNamesCount] = (char*)peerName;
    conf->peerNamesCount++;

    return TRUE;
}

static BOOLEAN
findDBPeerTitles(TI_Config */*conf*/, TI_DBEntry *dbEntry, char *peer)
{
    char **peerTitles = NULL;
    int peerTitleCount = 0;
    int i;

    /* discover all known AETitles for peer */
    peerTitleCount = CNF_aeTitlesForPeer(peer,
	&peerTitles);
    if (peerTitleCount <= 0) {
	errmsg("no AE titles defined (in: %s) for peer: %s", configFileName,
	    peer);
	return FALSE;	
    }

    /* check to make sure peer+AEtitle has access to database areas */
    for (i=0; i<peerTitleCount; i++) {
	if (CNF_peerInAETitle(dbEntry->title,
	    peerTitles[i], peer)) {

	    /* add peer title to database's peer title list */
	    if (dbEntry->peerTitles == NULL) {
		dbEntry->peerTitles = (char**) malloc(sizeof(char*));
	    } else {
		dbEntry->peerTitles = (char**) realloc(dbEntry->peerTitles,
		    (dbEntry->peerTitleCount + 1) * sizeof(char*));
	    }
	    dbEntry->peerTitles[dbEntry->peerTitleCount] = peerTitles[i];
	    dbEntry->peerTitleCount++;
	}
    }

    /* throw away the old list */
    free(peerTitles);

    return (dbEntry->peerTitleCount > 0);
}

static void
createConfigEntries(TI_Config *conf)
{
    char **ctnTitles = NULL;
    int ctnTitleCount = 0;
    int i, j;
    TI_DBEntry *dbEntry = NULL;

    conf->dbCount = 0;

    /* discover all the known CTN AE titles */
    ctnTitleCount = CNF_ctnTitles(&ctnTitles);

    for(i=0; i<ctnTitleCount; i++) {
	if (!TI_dbReadable(ctnTitles[i])) {
	    errmsg("Warning: db does not exist: %s", ctnTitles[i]);
	} else {
	    dbEntry = (TI_DBEntry*) malloc(sizeof(TI_DBEntry));
	    bzero((char*)dbEntry, sizeof(*dbEntry));
	    dbEntry->title = ctnTitles[i];

            for (j=0; j<conf->peerNamesCount; j++) {
                findDBPeerTitles(conf, dbEntry, conf->peerNames[j]);
            }

	    if (dbEntry->peerTitleCount > 0) {
	        /* add database to list, it is accessable by something */
	        if (conf->dbEntries == NULL) {
	            conf->dbEntries = 
		        (TI_DBEntry**) malloc(sizeof(TI_DBEntry*));
	        } else {
	            conf->dbEntries = (TI_DBEntry**) realloc(conf->dbEntries,
		        (conf->dbCount + 1) * sizeof(TI_DBEntry*));
	        }
	        conf->dbEntries[conf->dbCount] = dbEntry;
	        conf->dbCount++;
	    } else {
	        free(dbEntry);
	    }
	}
    }

    /* throw away the old lists */
    free(ctnTitles);

    /* add any remote DB entries */
    for (i=0; i<remoteDBTitlesCount; i++) {
        const char *peerName = NULL;
	int portNumber;
	if (CNF_peerForAETitle(remoteDBTitles[i], &peerName, &portNumber)) {

	    /* add DB */
	    dbEntry = (TI_DBEntry*) malloc(sizeof(TI_DBEntry));
	    bzero((char*)dbEntry, sizeof(*dbEntry));
	    dbEntry->title = remoteDBTitles[i];
	    dbEntry->isRemoteDB = TRUE;

	    if (conf->dbEntries == NULL) {
	        conf->dbEntries = 
		    (TI_DBEntry**) malloc(sizeof(TI_DBEntry*));
	    } else {
	        conf->dbEntries = (TI_DBEntry**) realloc(conf->dbEntries,
		    (conf->dbCount + 1) * sizeof(TI_DBEntry*));
	    }
	    conf->dbEntries[conf->dbCount] = dbEntry;
	    conf->dbCount++;

            for (j=0; j<conf->peerNamesCount; j++) {
    		char **peerTitles = NULL;
   		int peerTitleCount = 0;
    		int k;

                peerTitleCount = CNF_aeTitlesForPeer(conf->peerNames[j],
		    &peerTitles);
    		if (peerTitleCount <= 0) {
		    errmsg("no AE titles defined (in: %s) for peer: %s", 
		        configFileName, conf->peerNames[j]);
    		}

	    	for (k=0; k<peerTitleCount; k++) {
	    	    /* add peer title to database's peer title list */
	    	    if (dbEntry->peerTitles == NULL) {
		    	dbEntry->peerTitles = (char**) malloc(sizeof(char*));
	    	    } else {
			dbEntry->peerTitles = (char**) realloc(dbEntry->peerTitles,
		    	(dbEntry->peerTitleCount + 1) * sizeof(char*));
	    	    }
	    	    dbEntry->peerTitles[dbEntry->peerTitleCount] = peerTitles[k];
	    	    dbEntry->peerTitleCount++;
	    	}

    		/* throw away the old list */
    		free(peerTitles);

            }	    
	}
    }
}

/*
 * Handle interrupt signals.  
 * We only really need to make sure that the display is clear
 * before quiting.
 */
 
static void
signalHandler(int /*sig*/)
{
    TI_detatchAssociation(&conf, TRUE);

    exit(1);
}

/*
** CVS Log
** ti.cc,v
** Revision 1.5  1996/05/30 17:41:27  hewett
** Added command line argument (-DD) to explicitly disable communication
** with ctndisp and override the configuration file.
**
** Revision 1.4  1996/05/06 07:33:14  hewett
** Rearranged handing of display disabling code.
**
** Revision 1.3  1996/04/25 16:29:09  hewett
** Added char* parameter casts to bzero() calls.
**
** Revision 1.2  1996/04/22 16:33:03  hewett
** Removed calls to DSU_setTraceLevel().
**
** Revision 1.1  1996/04/22 10:27:25  hewett
** Initial release.
**
**
*/
