/*
**  Copyright (C) 1993/1994, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/
/*
** TI Network Routines
**
** Author: 	Andrew Hewett
**		Universitaet Oldenburg, Germany
** Created:	08/1993
**
** Last Update:		hewett
** Update Date:		1996/04/25 16:29:10
** Source File:		/share/dicom/cvs-depot/dcmtk/imagectn/apps/tinet.cc,v
** CVS/RCS Revision:	1.2
** Status:		Exp
**
** CVS/RCS Log at end of file
*/


#include "osconfig.h"    /* make sure OS specific configuration is included first */

#include <stdio.h>
#include <string.h>
#ifndef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <errno.h>

#include "ti.h"
#include "assoc.h"
#include "dimse.h"
#include "diutil.h"
#include "dcuid.h"
#include "dcfilefo.h"
#include "cnf.h"
#ifndef NODISPLAY
#include "dispuser.h"
#endif
#include "tiui.h"
#include "tiquery.h"
#include "tinet.h"

static char *abstractSyntaxes[] = {
    UID_VerificationSOPClass, 
    UID_FINDStudyRootQueryRetrieveInformationModel, 
    UID_ComputedRadiographyImageStorage,
    UID_CTImageStorage,
    UID_MRImageStorage,
    UID_NuclearMedicineImageStorage,
    UID_RETIRED_NuclearMedicineImageStorage,
    UID_UltrasoundImageStorage,
    UID_RETIRED_UltrasoundImageStorage,
    UID_UltrasoundMultiframeImageStorage,
    UID_RETIRED_UltrasoundMultiframeImageStorage,
    UID_XRayAngiographicImageStorage,
    UID_XRayAngiographicBiPlaneImageStorage,
    UID_XRayFluoroscopyImageStorage,
    UID_SecondaryCaptureImageStorage,
    UID_StandaloneOverlayStorage,
    UID_StandaloneCurveStorage,
    UID_StandaloneModalityLUTStorage,
    UID_StandaloneVOILUTStorage
};

BOOLEAN
TI_detatchAssociation(TI_Config *conf, BOOLEAN abortFlag)
{
    CONDITION cond;
    DIC_NODENAME presentationAddress;
    DIC_AE peerTitle;
    
    if (conf->assoc == NULL) {
    	return TRUE;	/* nothing to do */
    }

    ASC_getPresentationAddresses(conf->assoc->params, NULL, 
        presentationAddress);
    ASC_getAPTitles(conf->assoc->params, NULL, peerTitle, NULL);

    if (abortFlag) {
        /* abort association */
        if (verbose)
	    printf("Aborting Association (%s)\n", peerTitle);
        cond = ASC_abortAssociation(conf->assoc);
        if (!SUCCESS(cond)) {
	    errmsg("Association Abort Failed:");
	    COND_DumpConditions(); COND_PopCondition(TRUE);
        }
    } else {
        /* release association */
        if (verbose)
	    printf("Releasing Association (%s)\n", peerTitle);
        cond = ASC_releaseAssociation(conf->assoc);
        if (cond != ASC_NORMAL && cond != ASC_RELEASECONFIRMED) {
	    errmsg("Association Release Failed:");
	    COND_DumpConditions(); COND_PopCondition(TRUE);
        }
    }
    ASC_dropAssociation(conf->assoc);
    ASC_destroyAssociation(&conf->assoc);

    if (abortFlag) {
        printf("Aborted Association (%s,%s)\n", 
	    presentationAddress, peerTitle);
    } else {
        printf("Released Association (%s,%s)\n", 
	    presentationAddress, peerTitle);
    }

#ifndef NODISPLAY    
    DSU_stopAssociation(DSU_MainAssociation);
#endif
    
    return TRUE;
}



static CONDITION
addPresentationContexts(T_ASC_Parameters *params)
{
    CONDITION cond = ASC_NORMAL;
    int i;
    int pid = 1;

    /* 
    ** We prefer to accept Explicitly encoded transfer syntaxes.
    ** If we are running on a Little Endian machine we prefer 
    ** LittleEndianExplicitTransferSyntax to BigEndianTransferSyntax.
    ** Some SCP implementations will just select the first transfer
    ** syntax they support (this is not part of the standard) so
    ** organise the proposed transfer syntaxes to take advantage
    ** of such behaviour.
    */

    const char* transferSyntaxes[] = { 
	NULL, NULL, UID_LittleEndianImplicitTransferSyntax };

    /* gLocalByteOrder is defined in dcxfer.h */
    if (gLocalByteOrder == EBO_LittleEndian) {
	/* we are on a little endian machine */
	transferSyntaxes[0] = UID_LittleEndianExplicitTransferSyntax;
	transferSyntaxes[1] = UID_BigEndianExplicitTransferSyntax;
    } else {
	/* we are on a big endian machine */
	transferSyntaxes[0] = UID_BigEndianExplicitTransferSyntax;
	transferSyntaxes[1] = UID_LittleEndianExplicitTransferSyntax;
    }

    for (i=0; i<(int)DIM_OF(abstractSyntaxes) && SUCCESS(cond); i++) {
	cond = ASC_addPresentationContext(
	    params, pid, abstractSyntaxes[i],
	    transferSyntaxes, DIM_OF(transferSyntaxes));
	pid += 2;	/* only odd presentation context id's */
    }

    return cond;
}


BOOLEAN
TI_attachAssociation(TI_Config *conf)
{
    CONDITION cond;
    int port;
    const char *peer;
    DIC_NODENAME presentationAddress;
    T_ASC_Parameters *params;
    DIC_NODENAME localHost;
    DIC_AE myAETitle;

    if (conf->assoc != NULL) {
    	TI_detatchAssociation(conf, FALSE);
    }

    if (conf->dbEntries[conf->currentdb]->isRemoteDB) {
        strcpy(myAETitle, conf->myAETitle);
    } else {
        strcpy(myAETitle, conf->dbEntries[conf->currentdb]->title);
    }
    
    cond = ASC_createAssociationParameters(&params, conf->maxReceivePDULength);
    if (!SUCCESS(cond)) {
	errmsg("Help, cannot create association parameters:");
	COND_DumpConditions(); COND_PopCondition(TRUE);
	return FALSE;
    }
    ASC_setAPTitles(params, myAETitle, conf->currentPeerTitle, NULL);

    gethostname(localHost, sizeof(localHost) - 1);
    if (!CNF_peerForAETitle(conf->currentPeerTitle, &peer, &port)) {
	errmsg("Help, AE title (%s) no longer in config", 
	    conf->currentPeerTitle);
	ASC_destroyAssociationParameters(&params);
	return FALSE;
    }
    sprintf(presentationAddress, "%s:%d", peer, port);
    ASC_setPresentationAddresses(params, localHost, presentationAddress);

    cond = addPresentationContexts(params);
    if (!SUCCESS(cond)) {
	errmsg("Help, cannot add presentation contexts:");
	COND_DumpConditions();
	ASC_destroyAssociationParameters(&params);
	COND_PopCondition(TRUE);
	return FALSE;
    }
    if (debug) {
	printf("Request Parameters:\n");
	ASC_dumpParameters(params);
    }

    /* create association */
    if (verbose)
	printf("Requesting Association\n");
    cond = ASC_requestAssociation(conf->net, params, &conf->assoc);
    if (cond != ASC_NORMAL) {
	if (cond == ASC_ASSOCIATIONREJECTED) {
	    T_ASC_RejectParameters rej;

	    ASC_getRejectParameters(params, &rej);
	    errmsg("Association Rejected:");
	    ASC_printRejectParameters(stderr, &rej);
	    fprintf(stderr, "\n");
	    ASC_dropAssociation(conf->assoc);
	    ASC_destroyAssociation(&conf->assoc);
	    COND_PopCondition(TRUE);
	    return FALSE;
	} else {
	    errmsg("Association Request Failed: Peer (%s, %s)",
	        presentationAddress, conf->currentPeerTitle);
	    COND_DumpConditions();
	    ASC_dropAssociation(conf->assoc);
	    ASC_destroyAssociation(&conf->assoc);
	    COND_PopCondition(TRUE);
	    return FALSE;
	}
    }
    /* what has been accepted/refused ? */
    if (debug) {
	printf("Association Parameters Negotiated:\n");
	ASC_dumpParameters(params);
    }

    if (ASC_countAcceptedPresentationContexts(params) == 0) {
	errmsg("All Presentation Contexts Refused: Peer (%s, %s)",
	        presentationAddress, conf->currentPeerTitle);
	ASC_abortAssociation(conf->assoc);
	ASC_dropAssociation(conf->assoc);
	ASC_destroyAssociation(&conf->assoc);
	COND_PopCondition(TRUE);
	return FALSE;
    }

    if (verbose) {
	printf("Association Accepted (Max Send PDV: %lu)\n",
		conf->assoc->sendPDVLength);
    }

    printf("New Association Started (%s,%s)\n", presentationAddress,
        conf->currentPeerTitle);

#ifndef NODISPLAY	
    DSU_startAssociation(peer, conf->currentPeerTitle, myAETitle,
        DSU_Requestor, DSU_MainAssociation);
#endif
    
    return TRUE;
}

/*
 * Change Association
 */

BOOLEAN 
TI_changeAssociation(TI_Config *conf)
{
    DIC_AE actualPeerAETitle;
    BOOLEAN ok = TRUE;

    if (conf->assoc != NULL) {
        /* do we really need to change the association */
        ASC_getAPTitles(conf->assoc->params, NULL, actualPeerAETitle, NULL);
        if (strcmp(actualPeerAETitle, conf->currentPeerTitle) == 0) {
	    /* no need to change */
	    return TRUE;
	}
    }

    ok = TI_detatchAssociation(conf, FALSE);
    if (!ok) return ok;

    ok = TI_attachAssociation(conf);
    return ok;
}

BOOLEAN 
TI_sendEcho(TI_Config *conf)
{
    CONDITION cond = DIMSE_NORMAL;
    DIC_US msgId; 
    DIC_US status;
    DcmDataset *statusDetail = NULL;

    msgId = conf->assoc->nextMsgID++;
    printf("[MsgID %d] Echo, ", msgId);
    fflush(stdout);

    cond = DIMSE_echoUser(conf->assoc, msgId, DIMSE_BLOCKING, 0,
    	&status, &statusDetail);

    if (cond == DIMSE_NORMAL) {
	printf("Complete [Status: %s]\n",
	    DU_cstoreStatusString(status));
    } else {
        errmsg("Failed:");
	COND_DumpConditions();
	ASC_abortAssociation(conf->assoc);
	ASC_dropAssociation(conf->assoc);
	ASC_destroyAssociation(&conf->assoc);
	COND_PopCondition(TRUE);
#ifndef NODISPLAY
	/* inform display */
        DSU_stopAssociation(DSU_MainAssociation);
#endif
    }
    if (statusDetail != NULL) {
        printf("  Status Detail (should never be any):\n");
	statusDetail->print();
	delete statusDetail;
    }
    return (cond == DIMSE_NORMAL);
}


static void
storeProgressCallback(void */*callbackData*/, 
    T_DIMSE_StoreProgress *progress,
    T_DIMSE_C_StoreRQ */*req*/)
{
    int percent;
    static int lastPercent = 0;
    static int dotsSoFar = 0;
    int dotsToPrint;
    int i;

    switch (progress->state) {
    case DIMSE_StoreBegin:
	printf("  0%%________25%%_________50%%__________75%%________100%%\n");
	printf("  ");
	lastPercent = 0;
	dotsSoFar = 0;
	break;
    case DIMSE_StoreEnd:
	printf("\n");
#ifndef NODISPLAY
	if (displayTransferProgress) {
	    DSU_operationState(DSU_MainAssociation, DSU_StoreOperation,
		DSU_RunningOperation, DSU_Outgoing, 100);
	}
#endif
	break;
    default:
	if (progress->totalBytes == 0) {
	    percent = 100;
	} else {
	    percent = (int)(((float)progress->progressBytes / 
	        (float)progress->totalBytes) * 100.0);
	}
	dotsToPrint = (percent/2) - dotsSoFar;
	for (i=0; i<dotsToPrint; i++) {
	    putchar('-'); 
	    fflush(stdout);
	    dotsSoFar++; 
	}
	if (displayTransferProgress) {
	    if ((percent - lastPercent) >= 10) { /* 10% change */
		if (lastPercent < 100) {
#ifndef NODISPLAY
	            DSU_operationState(DSU_MainAssociation, DSU_StoreOperation,
		        DSU_RunningOperation, DSU_Outgoing, percent);
#endif
		}
	        lastPercent = percent;
	    }
	}
	break;
    }
}


BOOLEAN
TI_storeImage(TI_Config *conf, 
    char *sopClass, char *sopInstance, char * imgFile)
{
    CONDITION cond;
    DIC_US msgId;
    DcmDataset *statusDetail = NULL;
    T_ASC_PresentationContextID presId;
    T_DIMSE_C_StoreRQ req;
    T_DIMSE_C_StoreRSP rsp;
    DIC_PN patientsName; 
    DIC_CS studyId;
    DIC_IS seriesNumber;
    DIC_CS modality;
    DIC_IS imageNumber;
    int lockfd;
    
    if (strlen(sopClass) == 0) {
        printf("WARNING: CTN has deleted image, giving up (no sopClass): %s\n",
	       (imgFile)?(imgFile):("(nil)"));
        /* give up because if this image is gone, then others are also
	 * very likely to have disappeared.  The user should restart
	 * the operation when other activities have finished.
	 */
	return FALSE;
    }
    
    /* shared lock image file */
    lockfd = open(imgFile, O_RDONLY, 0666);
    if (lockfd < 0) {
        printf("WARNING: CTN has deleted image, giving up (no imgFile): %s\n",
	       (imgFile)?(imgFile):("(nil)"));
        /* give up because if this image is gone, then others are also
	 * very likely to have disappeared.  The user should restart
	 * the operation when other activities have finished.
	 */
	return FALSE;
    }
    flock(lockfd, LOCK_SH);

    /* which presentation context should be used */
    presId = ASC_findAcceptedPresentationContextID(conf->assoc, sopClass);
    if (presId == 0) {
	errmsg("No presentation context for: (%s) %s", 
	    DU_sopClassToModality(sopClass), sopClass);
	return FALSE;
    }

    TI_getInfoFromImage(imgFile, patientsName, studyId,
        seriesNumber, modality, imageNumber);

    /* start store */
    msgId = conf->assoc->nextMsgID++;
    printf("[MsgID %d] Store,\n", msgId);
    printf("  PatientsName: %s, StudyID: %s,\n", 
        patientsName, studyId);
    printf("  Series: %s, Modality: %s, Image: %s,\n",
        seriesNumber, modality, imageNumber);       
    printf("  Image UID: %s\n", sopInstance);
    fflush(stdout);
#ifndef NODISPLAY
    DSU_operationState(DSU_MainAssociation, DSU_StoreOperation,
        DSU_StartOperation, DSU_Outgoing, 0);
#endif
    bzero((char*)&req, sizeof(req));
    req.MessageID = msgId;
    strcpy(req.AffectedSOPClassUID, sopClass);
    strcpy(req.AffectedSOPInstanceUID, sopInstance);
    req.DataSetType = DIMSE_DATASET_PRESENT;
    req.Priority = DIMSE_PRIORITY_MEDIUM;

    cond = DIMSE_storeUser(conf->assoc, presId, &req,
       imgFile, NULL, storeProgressCallback, NULL, 
	DIMSE_BLOCKING, 0, 
	&rsp, &statusDetail);

    /* unlock image file */
    flock(lockfd, LOCK_UN);
    close(lockfd);

    if (cond == DIMSE_NORMAL) {
	printf("[MsgID %d] Complete [Status: %s]\n", msgId,
	    DU_cstoreStatusString(rsp.Status));
    } else {
        errmsg("[MsgID %d] Failed:", msgId);
	COND_DumpConditions();
	ASC_abortAssociation(conf->assoc);
	ASC_dropAssociation(conf->assoc);
	ASC_destroyAssociation(&conf->assoc);
	COND_PopCondition(TRUE);
#ifndef NODISPLAY
	/* inform display */
        DSU_stopAssociation(DSU_MainAssociation);
#endif
    }
    if (statusDetail != NULL) {
        printf("  Status Detail:\n");
	statusDetail->print();
	delete statusDetail;
    }

#ifndef NODISPLAY
    if (cond == DIMSE_NORMAL) {    
        DSU_operationState(DSU_MainAssociation, DSU_StoreOperation,
            DSU_StopOperation, DSU_Outgoing, 100);
    }
#endif
    
    return (cond == DIMSE_NORMAL);
}

 
/*
 * Find for remote DBs
 */  

typedef struct {
    TI_GenericEntryCallbackFunction cbf;
    TI_GenericCallbackStruct *cbs;
} TI_LocalFindCallbackData;

static void
findCallback(
	/* in */
	void *callbackData, 
	T_DIMSE_C_FindRQ */*request*/, /* original find request */
	int responseCount, 
	T_DIMSE_C_FindRSP *response,	/* pending response received */
	DcmDataset *responseIdentifiers /* pending response identifiers */
	)
{
    TI_LocalFindCallbackData *cbd;
    
    cbd = (TI_LocalFindCallbackData*)callbackData;
    
    if (verbose) {
        printf("Find Response %d:\n", responseCount);
        DIMSE_printCFindRSP(stdout, response);
        printf("Identifiers %d:\n", responseCount);
	responseIdentifiers->print();
    }

    /* call the callback function */
    cbd->cbf(cbd->cbs, responseIdentifiers);

    /* responseIdentifers will be deleted in DIMSE_findUser() */
}
 
BOOLEAN
TI_remoteFindQuery(TI_Config *conf, TI_DBEntry *db, DcmDataset *query,
    TI_GenericEntryCallbackFunction callbackFunction,
    TI_GenericCallbackStruct *callbackData)
{
    BOOLEAN ok = TRUE;
    TI_LocalFindCallbackData cbd;
    CONDITION           cond;
    T_ASC_PresentationContextID presId;
    DIC_US      	msgId;
    T_DIMSE_C_FindRQ	req;
    T_DIMSE_C_FindRSP	rsp;
    DcmDataset		*statusDetail = NULL;

    conf->currentPeerTitle = db->title;

    /* make sure we have an association */     
    ok = TI_changeAssociation(conf);
    if (!ok) return FALSE;

    cbd.cbf = callbackFunction;
    cbd.cbs = callbackData;

    /* which presentation context should be used */
    presId = ASC_findAcceptedPresentationContextID(conf->assoc, 
        UID_FINDStudyRootQueryRetrieveInformationModel);
    if (presId == 0) {
        errmsg("No Presentation Context for Find Operation");
	return FALSE;
    }

    msgId =  conf->assoc->nextMsgID++;

    if (verbose) {
	printf("Sending Find SCU RQ: MsgID %d:\n", msgId);
	query->print();
    }

    req.MessageID = msgId;
    strcpy(req.AffectedSOPClassUID, 
	   UID_FINDStudyRootQueryRetrieveInformationModel);
    req.Priority = DIMSE_PRIORITY_LOW;
    
    cond = DIMSE_findUser(conf->assoc, presId, &req, query,
    	findCallback, &cbd, DIMSE_BLOCKING, 0, &rsp, &statusDetail);

    if (cond == DIMSE_NORMAL) {
        if (verbose) {
	    DIMSE_printCFindRSP(stdout, &rsp);
        }
    } else {
        errmsg("Find Failed:");
	COND_DumpConditions();
    }
    if (statusDetail != NULL) {
        printf("  Status Detail:\n");
	statusDetail->print();
	delete statusDetail;
    }

    return SUCCESS(cond);

}

/*
** CVS Log
** tinet.cc,v
** Revision 1.2  1996/04/25 16:29:10  hewett
** Added char* parameter casts to bzero() calls.
**
** Revision 1.1  1996/04/22 10:27:26  hewett
** Initial release.
**
**
*/
